#! /usr/bin/python
import sys

from edge import eventlog
import setupstore.actions.pipeline.speed_pipeline_creator as pipeline_creator
import setupstore.actions.pipeline.ui_pipeline as pipeline_deleter
import setupstore.actions.pipeline.ui_pipeline_set as ui_pipeline_set
from setupstore.actions.pipeline.edge_creator import *
from edge.path_helpers import *
import edge.pipeline_helpers as helpers


class _ArgSet:
    ACTION_CREATE = 'create'
    ACTION_REMOVE = 'remove'
    ACTION_REGENERATE = 'regenerate'

    _CREATE_ACTION_VENDOR = '--vendor'
    _CREATE_ACTION_MODEL = '--model'
    _CREATE_ACTION_SERIAL = '--serial'

    _REMOVE_ACTION_PID = '--pid'

    _ARG_VALUE_SEPARATOR = '='

    def __init__(self, action, create_vendor=None, create_model=None, create_serial=None, remove_pid=None):
        """
        :type action: str
        :type create_vendor: str | None
        :type create_model: str | None
        :type create_serial: str | None
        :type remove_pid: str | None
        """
        self.action = action
        self.create_vendor = create_vendor
        self.create_model = create_model
        self.create_serial = create_serial
        self.remove_pid = remove_pid

    @staticmethod
    def show_usage():
        print 'pipeline-ctl create %s%s<vendor> %s%s<model> %s%s<serial>' \
              % (_ArgSet._CREATE_ACTION_VENDOR, _ArgSet._ARG_VALUE_SEPARATOR,
                 _ArgSet._CREATE_ACTION_MODEL, _ArgSet._ARG_VALUE_SEPARATOR,
                 _ArgSet._CREATE_ACTION_SERIAL, _ArgSet._ARG_VALUE_SEPARATOR)

        print 'pipeline-ctl remove %s%s<pipeline-id>' \
              % (_ArgSet._REMOVE_ACTION_PID, _ArgSet._ARG_VALUE_SEPARATOR)

        print 'pipeline-ctl regenerate'

    @staticmethod
    def _deserialize_create(raw_arg_set):
        """
        :type raw_arg_set: list of str
        :rtype _ArgSet
        """
        if len(raw_arg_set) != 3:
            raise ValueError('invalid number of arguments')

        vendor = None
        model = None
        serial = None
        for raw_arg in raw_arg_set:
            raw_name_value = raw_arg.split(_ArgSet._ARG_VALUE_SEPARATOR, 1)
            if len(raw_name_value) != 2:
                raise ValueError('invalid argument %s' % raw_arg)

            raw_name, raw_value = raw_name_value
            raw_name = raw_name.strip()
            raw_value = raw_value.strip()

            if raw_name == _ArgSet._CREATE_ACTION_VENDOR:
                vendor = raw_value
                continue

            if raw_name == _ArgSet._CREATE_ACTION_SERIAL:
                serial = raw_value
                continue

            if raw_name == _ArgSet._CREATE_ACTION_MODEL:
                model = raw_value
                continue

            raise ValueError('invalid argument %s' % raw_arg)

        if not reduce(lambda x, y: x and y, [vendor, model, serial]):
            raise ValueError('invalid number of arguments')

        return _ArgSet(_ArgSet.ACTION_CREATE, create_vendor=vendor, create_model=model, create_serial=serial)

    @staticmethod
    def _deserialize_remove(raw_arg_set):
        """
        :type raw_arg_set: list of str
        :rtype _ArgSet
        """
        if len(raw_arg_set) != 1:
            raise ValueError('invalid number of arguments')

        raw_name_value = str(raw_arg_set[0]).split(_ArgSet._ARG_VALUE_SEPARATOR, 1)
        if len(raw_name_value) != 2:
            raise ValueError('invalid argument')

        raw_name, raw_value = raw_name_value
        raw_name = raw_name.strip()
        raw_value = raw_value.strip()

        if raw_name != _ArgSet._REMOVE_ACTION_PID:
            raise ValueError('invalid argument')

        return _ArgSet(_ArgSet.ACTION_REMOVE, remove_pid=raw_value)

    @staticmethod
    def _deserialize_regenerate(raw_arg_set):
        """
        :type raw_arg_set: list of str
        :rtype _ArgSet
        """
        if len(raw_arg_set) != 0:
            raise ValueError('invalid number of arguments')

        return _ArgSet(_ArgSet.ACTION_REGENERATE)

    @staticmethod
    def deserialize(raw_arg_set):
        """
        :type raw_arg_set: list of str
        :rtype _ArgSet
        """
        raw_arg_count = len(raw_arg_set)
        if raw_arg_count < 2:
            raise ValueError('invalid number of arguments')

        raw_action = str(raw_arg_set[1])

        if raw_action == _ArgSet.ACTION_CREATE:
            return _ArgSet._deserialize_create(raw_arg_set[2:])

        elif raw_action == _ArgSet.ACTION_REMOVE:
            return _ArgSet._deserialize_remove(raw_arg_set[2:])

        elif raw_action == _ArgSet.ACTION_REGENERATE:
            return _ArgSet._deserialize_regenerate(raw_arg_set[2:])

        raise ValueError('invalid command-line')


def _handle_regenerate():
    pipeline_jnode_set = helpers.get_pipeline_jnode_set2()
    if pipeline_jnode_set is None:
        eventlog.log_d("fail: pipeline-jnode-set is absent")
        return

    for pipeline_jnode in pipeline_jnode_set:
        ui_pid = pipeline_jnode['id']
        ui_ptype = pipeline_jnode['type']

        eventlog.log_d('try: pipeline-%s' % ui_pid)
        # noinspection PyBroadException
        try:
            # create edge-jnode-set
            generate(ui_pid, UiIdentity(ui_ptype, '', ''), None)
            eventlog.log_d('done: pipeline-%s' % ui_pid)
            continue

        except NotImplementedError:
            eventlog.log_d('fail: pipeline-%s (camera not supported)' % ui_pid)

        except LockError:
            eventlog.log_d('fail: pipeline-%s (lock error)' % ui_pid)

        except:
            eventlog.log_d('fail: pipeline-%s (traceback)' % ui_pid)
            eventlog.log_x()

    eventlog.log_d('regeneration config done')


def _main(raw_arg_set):
    arg_set = None
    # noinspection PyBroadException
    try:
        arg_set = _ArgSet.deserialize(raw_arg_set)
    except ValueError:
        eventlog.log_x()
        _ArgSet.show_usage()
        exit(1)
    except:
        eventlog.log_x()
        exit(1)

    if arg_set.action == _ArgSet.ACTION_CREATE:
        pipeline_creator.create_speed_pipeline(
            arg_set.create_vendor, arg_set.create_model, arg_set.create_serial)
        helpers.sighup_setupstore()
        helpers.restart_webserver_lazy()

    elif arg_set.action == _ArgSet.ACTION_REMOVE:
        pipeline_deleter.delete_pipeline(arg_set.remove_pid)
        helpers.sighup_setupstore()
        helpers.restart_webserver_lazy()

    elif arg_set.action == _ArgSet.ACTION_REGENERATE:
        _handle_regenerate()

    exit(0)


if __name__ == '__main__':
    eventlog.open_log('pipeline-ctl', 'debug')
    try:
        _main(sys.argv)
    finally:
        eventlog.close_log()
