import os

from edge import eventlog
from edge.admin.reply_helpers import Reply, REPLY_SUCCESS_200, REPLY_NOT_FOUND_404
from edge.admin.cgi import (run_cgi, CgiModule, CgiPathInfo, CgiRequest,
                            CgiBadRequestError, CgiNotFoundError,
                            CgiConflictError, CgiInternalServerError)


from webserver.db.wanted.api import (init as init_wanted, fini as fini_wanted,
                                     WantedPack, WantedError)

_PATH_PROFILES = 'profiles'

_PATH_LISTS = 'list'

_LOAD_ACTION_NONE = 'none'
_LOAD_ACTION_RM = 'rm'
_LOAD_ACTION_CP = 'cp'

_API_TO_DB_ACTION_MAP = {
    _LOAD_ACTION_CP: 'set',
    _LOAD_ACTION_RM: 'unset',
    _LOAD_ACTION_NONE: 'none'
}


def _get_time(qs, key_to_find):
    """
    :type qs: str
    :type key_to_find: str
    :rtype: str
    """
    import urlparse
    import iso8601
    url_query = urlparse.parse_qs(qs[1:], strict_parsing=True)

    iso_datetime = None
    for key, values in url_query.iteritems():
        # deserialize datetime-max
        if key == key_to_find:
            if len(values) != 1:
                eventlog.log_i('fail: hppt400')
                raise CgiNotFoundError

            # url parse replases '+' to ' '
            raw_iso_datetime = values.pop()
            iso_datetime = raw_iso_datetime.replace(' ', '+')
            try:
                iso8601.parse_date(iso_datetime)
            except iso8601.ParseError as x:
                eventlog.log_i('fail: hppt400')
                raise CgiNotFoundError

            continue

    return iso_datetime


class _PathInfo(CgiPathInfo):
    def __init__(self, *args, **kwargs):
        CgiPathInfo.__init__(self, *args, **kwargs)

    @staticmethod
    def unpack(raw_path_info):
        """
        :type raw_path_info: str
        :rtype: _PathInfo
        """
        # eventlog.log_d('unpack %s' % type(raw_path_info))
        return _PathInfo(raw_path_info.split('/')[1:])


class _Module(CgiModule):
    """
    GET     /api/v2/sys/db/wanted
    PUT     /api/v2/sys/db/wanted  /list/:owner/:name
    DELETE  /api/v2/sys/db/wanted  /lists/:owner/:name

    GET     /api/v2/sys/db/wanted  /profiles/:name
    PUT     /api/v2/sys/db/wanted  /profiles/:name
    DELETE  /api/v2/sys/db/wanted  /profiles/:name
    DELETE  /api/v2/sys/db/wanted  /profiles/:name/log
    """
    def __init__(self, handler):
        """
        :type handler: () -> Reply
        """
        CgiModule.__init__(self, handler)

    @classmethod
    def unpack(cls, request, wantedpack):
        """
        :type request: CgiRequest
        :type wantedpack: WantedPack
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        method_to_handlers_map = {
            'GET': [
                cls._get,
                cls._get_profile
            ],
            'POST': [],
            'PUT': [
                cls._put_wanted_list,
                cls._put_profile
            ],
            'DELETE': [
                cls._delete_wanted_list,
                cls._delete_profile,
                cls._cleanup,
                cls._rollback
            ],
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request, wantedpack)

            except CgiBadRequestError:
                raise

            except CgiNotFoundError:
                pass

        eventlog.log_i('fail: http404')
        raise CgiNotFoundError

    @staticmethod
    def _get(request, wantedpack):
        """
        :type request: CgiRequest
        :type wantedpack: WantedPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 0:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            reply_jnode = wantedpack.get()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_wanted_list(request, wantedpack):
        """
        :type request: CgiRequest
        :type wantedpack: WantedPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 3:
                raise CgiNotFoundError

            if path_info[0] != _PATH_LISTS:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if '*' in request.path_info[1] or '*' in request.path_info[2]:
                eventlog.log_i('fail: hppt400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            try:
                reply_jnode = wantedpack.delete_watch_list(request.path_info[1], request.path_info[2])
            except WantedError:
                raise CgiNotFoundError

            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _put_wanted_list(request, wantedpack):
        """
        :type request: CgiRequest
        :type wantedpack: WantedPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 1:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'list':
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        list_owner = os.environ.get('LIST_OWNER')
        list_name = os.environ.get('LIST_NAME')
        list_path = os.environ.get('LIST_PATH')
        prof_name = os.environ.get('PROF_NAME', 'default')
        comment = os.environ.get('LIST_COMMENT')

        prof_same = os.environ.get('PROF_SAME', _LOAD_ACTION_NONE)
        prof_new = os.environ.get('PROF_NEW', _LOAD_ACTION_NONE)
        prof_old = os.environ.get('PROF_OLD', _LOAD_ACTION_NONE)

        if prof_same not in [_LOAD_ACTION_CP, _LOAD_ACTION_RM, _LOAD_ACTION_NONE]:
            eventlog.log_i('fail: 400')
            raise CgiBadRequestError

        if prof_new not in [_LOAD_ACTION_CP, _LOAD_ACTION_NONE]:
            eventlog.log_i('fail: 400')
            raise CgiBadRequestError

        if prof_old not in [_LOAD_ACTION_RM, _LOAD_ACTION_NONE]:
            eventlog.log_i('fail: 400')
            raise CgiBadRequestError

        prof_same = _API_TO_DB_ACTION_MAP[prof_same]
        prof_new = _API_TO_DB_ACTION_MAP[prof_new]
        prof_old = _API_TO_DB_ACTION_MAP[prof_old]

        if None in [list_owner, list_name, list_path]:
            raise CgiBadRequestError

        def handler():
            """
            :rtype: Reply
            """
            reply_jnode = {}
            try:
                wantedpack.load_watch_list(list_owner, list_name, list_path, prof_name, comment, prof_same, prof_new, prof_old)
            except WantedError as x:
                raise CgiBadRequestError

            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_profile(request, wantedpack):
        """
        :type request: CgiRequest
        :type wantedpack: WantedPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 2:
                raise CgiNotFoundError

            if path_info[0] != _PATH_PROFILES:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            try:
                reply_jnode = wantedpack.drop_profile(request.path_info[1])
            except WantedError:
                raise CgiNotFoundError

            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _cleanup(request, wantedpack):
        """
        :type request: CgiRequest
        :type wantedpack: WantedPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 4:
                raise CgiNotFoundError

            if path_info[0] != _PATH_LISTS:
                raise CgiNotFoundError

            if path_info[3] != 'log':
                raise CgiNotFoundError

        def check_args():
            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.query_string is None:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if not request.query_string.startswith('?'):
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if '*' in request.path_info[1] or '*' in request.path_info[2]:
                eventlog.log_i('fail: hppt400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        max_time = _get_time(request.query_string, 'max-time')
        if max_time is None:
            raise CgiNotFoundError

        def handler():
            """
            :rtype: Reply
            """
            rc = REPLY_SUCCESS_200
            try:
                wantedpack.cleanup_list(request.path_info[1], request.path_info[2], max_time)
            except WantedError:
                rc = REPLY_NOT_FOUND_404

            jnode = wantedpack.get()
            return Reply(status=rc, body=jnode)
        # handler()

        return _Module(handler)

    @staticmethod
    def _rollback(request, wantedpack):
        """
        :type request: CgiRequest
        :type wantedpack: WantedPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 4:
                raise CgiNotFoundError

            if path_info[0] != _PATH_LISTS:
                raise CgiNotFoundError

            if path_info[3] != 'log':
                raise CgiNotFoundError

        def check_args():
            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.query_string is None:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if not request.query_string.startswith('?'):
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if '*' in request.path_info[1] or '*' in request.path_info[2]:
                eventlog.log_i('fail: hppt400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        min_time = _get_time(request.query_string, 'min-time')
        if min_time is None:
            raise CgiNotFoundError

        def handler():
            """
            :rtype: Reply
            """
            rc = REPLY_SUCCESS_200
            try:
                wantedpack.revert_list(request.path_info[1], request.path_info[2], min_time)
            except WantedError:
                rc = REPLY_NOT_FOUND_404

            jnode = wantedpack.get()
            return Reply(status=rc, body=jnode)

        return _Module(handler)

    @staticmethod
    def _get_profile(request, wantedpack):
        """
        :type request: CgiRequest
        :type wantedpack: WantedPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 2:
                raise CgiNotFoundError

            if path_info[0] != _PATH_PROFILES:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            try:
                reply_jnode = wantedpack.load_profile(request.path_info[1])
            except WantedError:
                raise CgiNotFoundError

            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _put_profile(request, wantedpack):
        """
        :type request: CgiRequest
        :type wantedpack: WantedPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 2:
                raise CgiNotFoundError

            if path_info[0] != _PATH_PROFILES:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body is None:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            try:
                reply_jnode = wantedpack.save_profile(
                    request.path_info[1], request.body_jnode['args']['template'])
            except WantedError:
                raise CgiNotFoundError

            return Reply(body=reply_jnode)

        return _Module(handler)


def _main():
    wanted_pack = init_wanted()

    try:
        run_cgi(
            lambda arg: _PathInfo.unpack(arg),
            lambda arg: _Module.unpack(arg, wanted_pack))
    finally:
        fini_wanted(wanted_pack)


if __name__ == '__main__':
    eventlog.open_log('cgi-sys-db-wanted', 'debug')
    eventlog.log_i('try')
    try:
        _main()
    finally:
        eventlog.close_log()
