import os
import iso8601

from edge import eventlog
from edge.admin.reply_helpers import Reply
from edge.admin.cgi import (run_cgi, CgiModule, CgiPathInfo, CgiRequest,
                            CgiBadRequestError, CgiNotFoundError,
                            CgiConflictError, CgiInternalServerError)


from webserver.sys.test.api import (init as init_testpack,
                                    fini as fini_testpack, TestPack)
from webserver.sys.test.test import TestError, TestArgsError, TestLimitsError


class _PathInfo(CgiPathInfo):
    def __init__(self, *args, **kwargs):
        CgiPathInfo.__init__(self, *args, **kwargs)

    @property
    def session_ref(self):
        return self[1]

    @property
    def test_type(self):
        return self[3]

    @property
    def test_run(self):
        return self[4]

    @property
    def test_ref(self):
        return self[3], self[4]

    @property
    def test_step_ref(self):
        return self[6]

    @staticmethod
    def unpack(raw_path_info):
        """
        :type raw_path_info: str
        :rtype: _PathInfo
        """
        # eventlog.log_d('unpack %s' % type(raw_path_info))
        return _PathInfo(raw_path_info.split('/')[1:])


class _Module(CgiModule):
    """
    GET     /api/v2/sys/test
    PUT     /api/v2/sys/test  /results
    POST    /api/v2/sys/test  /results/report
    DELETE  /api/v2/sys/test  /results/report/:ref
    POST    /api/v2/sys/test  /sessions
    PUT     /api/v2/sys/test  /sessions/:session
    DELETE  /api/v2/sys/test  /sessions/:session

    POST    /api/v2/sys/test  /sessions/:session/tests
    GET     /api/v2/sys/test  /sessions/:session/tests/:test/:run
    PUT     /api/v2/sys/test  /sessions/:session/tests/:test/:run
    DELETE  /api/v2/sys/test  /sessions/:session/tests/:test/:run

    POST    /api/v2/sys/test  /sessions/:session/tests/:test/:run/log
    PUT     /api/v2/sys/test  /sessions/:session/tests/:test/:run/log/all
    PUT     /api/v2/sys/test  /sessions/:session/tests/:test/:run/log/:item
    DELETE  /api/v2/sys/test  /sessions/:session/tests/:test/:run/log
    DELETE  /api/v2/sys/test  /sessions/:session/tests/:test/:run/log/:item
    """
    def __init__(self, handler):
        """
        :type handler: () -> Reply
        """
        CgiModule.__init__(self, handler)

    @classmethod
    def unpack(cls, request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        method_to_handlers_map = {
            'GET': [
                cls._get,
                cls._get_sessions_session_tests_test_run
            ],
            'POST': [
                cls._post_sessions,
                cls._post_report,
                cls._post_sessions_session_tests,
                cls._post_sessions_session_tests_test_run_log
            ],
            'PUT': [
                cls._put_sessions_session,
                cls._put_test_results,
                cls._put_sessions_session_tests_test_run,
                cls._put_sessions_session_tests_test_run_log_all,
                cls._put_sessions_session_tests_test_run_log_item
            ],
            'DELETE': [
                cls._delete_report,
                cls._delete_sessions_session,
                cls._delete_sessions_session_tests_test_run,
                cls._delete_sessions_session_tests_test_run_log_item,
                cls._delete_sessions_session_tests_test_run_log
            ],
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request, testpack)

            except CgiBadRequestError:
                raise

            except CgiNotFoundError:
                pass

        eventlog.log_i('fail: http404')
        raise CgiNotFoundError

    @staticmethod
    def _get(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 0:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            reply_jnode = testpack.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _post_report(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 2:
                raise CgiNotFoundError

            if path_info[0] != 'results':
                raise CgiNotFoundError

            if path_info[1] != 'report':
                raise CgiNotFoundError

        def check_args():
            """
            :rtype: str
            """
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            try:
                tmp_path = os.environ['BLOB_PATH']
            except KeyError:
                eventlog.log_e('fail: http500 (env.BLOB_PATH not found)')
                raise CgiInternalServerError

            return tmp_path

        check_path_info(request.path_info)
        blob_filepath = check_args()

        def handler():
            """
            :rtype: Reply
            """
            testpack.add_report(blob_filepath)

            reply_jnode = testpack.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_report(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 3:
                raise CgiNotFoundError

            if path_info[0] != 'results':
                raise CgiNotFoundError

            if path_info[1] != 'report':
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            try:
                testpack.delete_report(request.path_info[2])

                reply_jnode = testpack.pack()
                return Reply(body=reply_jnode)

            except TestError:
                eventlog.log_e('fail: delete report')
                raise CgiInternalServerError

        return _Module(handler)

    @staticmethod
    def _post_sessions(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 1:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        auto_tests = False
        if request.body:
            auto_tests = request.body_jnode.get('tests', False)
            if type(auto_tests) != bool:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        def handler():
            """
            :rtype: Reply
            """
            try:
                session = testpack.add_session()
                if auto_tests:
                    session.add_tests()

                reply_jnode = testpack.pack()
                return Reply(body=reply_jnode)

            except TestLimitsError:
                eventlog.log_i('fail: http409')
                raise CgiConflictError

            except TestError:
                eventlog.log_e('fail: add session')
                raise CgiInternalServerError

        return _Module(handler)

    @staticmethod
    def _put_sessions_session(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 2:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body is None:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                eventlog.log_i('fail: http404')
                raise CgiNotFoundError

            session.set_summary(request.body_jnode)

            reply_jnode = testpack.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _put_test_results(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 1:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'results':
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body is not None:
                certificate_jnode = request.body_jnode.get('certificate')
                if certificate_jnode:
                    ref_number = certificate_jnode.get('ref-number')
                    max_time = certificate_jnode.get('max-time')
                    try:
                        iso8601.parse_date(max_time)

                    except iso8601.ParseError as x:
                        eventlog.log_i('fail: <{}>'.format(x.message))
                        raise CgiBadRequestError

                    if None in [ref_number, max_time]:
                        raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            testpack.add_results(request.body_jnode if request.body else None)
            reply_jnode = testpack.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_sessions_session(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 2:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            path_info = request.path_info  # type: _PathInfo
            try:
                if not testpack.del_session(path_info.session_ref):
                    eventlog.log_i("fail: http404")
                    raise CgiNotFoundError

            except TestError:
                eventlog.log_e('fail: del session')
                raise CgiInternalServerError

            reply_jnode = testpack.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _post_sessions_session_tests(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 3:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise CgiNotFoundError

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body is None:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                eventlog.log_i('fail: http404')
                raise CgiNotFoundError

            session.add_test(request.body_jnode)

            reply_jnode = testpack.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _get_sessions_session_tests_test_run(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 5:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise CgiNotFoundError

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)

            test = session.get_test(path_info.test_ref)
            if test is None:
                eventlog.log_i('fail: http404')
                raise CgiNotFoundError

            reply_jnode = test.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _put_sessions_session_tests_test_run(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 5:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise CgiNotFoundError

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            body_jnode = request.body_jnode if request.body else None
            if body_jnode:
                summary_jnode = body_jnode.get('summary') or dict()
                if summary_jnode.get('status') not in ['failed', 'passed']:
                    eventlog.log_i('fail: http400')
                    raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)

            test = session.get_test(path_info.test_ref)
            if test is None:
                eventlog.log_i('fail: http404')
                raise CgiNotFoundError

            test.set_summary(request.body_jnode if request.body else None)
            reply_jnode = test.pack()

            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_sessions_session_tests_test_run(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 5:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise CgiNotFoundError

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            session.del_test(path_info.test_ref)

            reply_jnode = testpack.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _post_sessions_session_tests_test_run_log(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 6:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise CgiNotFoundError

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run/log
            if path_info[5] != 'log':
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            test = session.get_test(path_info.test_ref)

            if test is None:
                eventlog.log_i('fail: http404')
                raise CgiNotFoundError

            try:
                test.add_step({
                    'args': request.body_jnode if request.body else None
                })

            except TestArgsError:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            except TestLimitsError:
                eventlog.log_i('fail: http409')
                raise CgiConflictError

            except TestError:
                eventlog.log_i('fail: http500')
                raise CgiInternalServerError

            reply_jnode = test.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _put_sessions_session_tests_test_run_log_item(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 7:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise CgiNotFoundError

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run/log
            if path_info[5] != 'log':
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run/log/:item
            test_step = test.get_step(path_info.test_step_ref)
            if test_step is None:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body is None:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

            step = test.get_step(path_info.test_step_ref)
            if step is None:
                raise CgiNotFoundError

            step.edit(request.body_jnode)

            reply_jnode = test.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _put_sessions_session_tests_test_run_log_all(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 7:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise CgiNotFoundError

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run/log
            if path_info[5] != 'log':
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run/log/all
            if path_info.test_step_ref != 'all':
                raise CgiNotFoundError

        report = False

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            body_jnode = request.body_jnode if request.body else dict()
            add_report = body_jnode.get('report', False)
            return add_report

        check_path_info(request.path_info)
        report = check_args()

        def handler():
            """
            :rtype: Reply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

            if not test.set_steps_report(report):
                raise CgiNotFoundError

            reply_jnode = test.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_sessions_session_tests_test_run_log_item(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 7:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise CgiNotFoundError

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run/log
            if path_info[5] != 'log':
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run/log/:item
            test_step = test.get_step(path_info.test_step_ref)
            if test_step is None:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)

            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

            if not test.del_step(path_info.test_step_ref):
                raise CgiNotFoundError

            reply_jnode = test.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _delete_sessions_session_tests_test_run_log(request, testpack):
        """
        :type request: CgiRequest
        :type testpack: TestPack
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 6:
                raise CgiNotFoundError

            # /sessions
            if path_info[0] != 'sessions':
                raise CgiNotFoundError

            # /sessions/:session
            session = testpack.get_session(path_info.session_ref)
            if session is None:
                raise CgiNotFoundError

            # /sessions/:session/tests
            if path_info[2] != 'tests':
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

            # /sessions/:session/tests/:test/:run/log
            if path_info[5] != 'log':
                raise CgiNotFoundError

        def unpack_args():
            """
            :rtype: list[str]
            """
            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if not request.query_string:
                return []

            if not request.query_string.startswith('?'):
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            from collections import defaultdict
            name_to_value = defaultdict(list)
            for name_value in request.query_string[1:].split('&'):
                name, value = name_value.split('=', 1)
                name_to_value[name] += [value]

            if len(name_to_value) != 1:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            items = name_to_value.get('item', [])
            if not items:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            return items

        check_path_info(request.path_info)
        step_refs = unpack_args()

        def handler():
            """
            :rtype: Reply
            """
            path_info = request.path_info  # type: _PathInfo
            session = testpack.get_session(path_info.session_ref)
            test = session.get_test(path_info.test_ref)
            if test is None:
                raise CgiNotFoundError

            if not test.del_steps(step_refs):
                raise CgiNotFoundError

            reply_jnode = test.pack()
            return Reply(body=reply_jnode)

        return _Module(handler)


def _main():
    testpack = init_testpack()
    try:
        run_cgi(
            lambda arg: _PathInfo.unpack(arg),
            lambda arg: _Module.unpack(arg, testpack))
    finally:
        fini_testpack(testpack)


if __name__ == '__main__':
    eventlog.open_log('cgi-sys-test', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
