import ujson
import os

from edge.admin.ui_types import BaseCreator
from edge.path_helpers import *
from os.path import join as join_path

import edge.pipeline_helpers as helpers


def rc_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules, avaliable_ui_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :type avaliable_ui_modules: set of UiIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None

    # mv
    if ui_entry_point in [Pipeline.Mv.Ui.VIDEOSOURCE_0]:

        used_ui_modules.add(Pipeline.Mv.Ui.VIDEOSOURCE_0)
        used_edge_modules.add(Pipeline.Mv.Edge.RAW_CODEC_0)
        creator_ctor = RcCreator.create

    elif ui_entry_point in [Pipeline.Mv.Ui.PIPELINE]:
        used_ui_modules.add(Pipeline.Mv.Ui.VIDEOSOURCE_0)
        creator_ctor = RcCreator.create

    # speed
    elif ui_entry_point in [Pipeline.Speed.Ui.VIDEOSOURCE_0]:

        used_ui_modules.add(Pipeline.Speed.Ui.VIDEOSOURCE_0)
        used_edge_modules.add(Pipeline.Speed.Edge.RAW_CODEC_0)
        creator_ctor = RcCreator.create

    elif ui_entry_point in [Pipeline.Speed.Ui.PIPELINE]:
        used_ui_modules.add(Pipeline.Speed.Ui.VIDEOSOURCE_0)
        creator_ctor = RcCreator.create

    return creator_ctor


class RcCreator(BaseCreator):
    def __init__(self, id_main_host_setupstore):
        """
        :type id_main_host_setupstore: str, dict, dict, str
        """
        BaseCreator.__init__(self)
        self._id_main_host_setupstore = id_main_host_setupstore

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: MsCreator
        """
        # mv
        if ui_entry_point in [Pipeline.Mv.Ui.PIPELINE, Pipeline.Mv.Ui.VIDEOSOURCE_0]:

            id_main_setupstore = RcCreator._serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map)
            creator = RcCreator(id_main_setupstore)

        # speed
        elif ui_entry_point in [Pipeline.Speed.Ui.PIPELINE, Pipeline.Speed.Ui.VIDEOSOURCE_0]:

            id_main_setupstore = RcCreator._serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map)
            creator = RcCreator(id_main_setupstore)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: str, dict, dict, str
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point.ptype == PTYPE_LPR_MV:
            ui_videosource_jnode = ui_identity_to_jnode_map[Pipeline.Mv.Ui.VIDEOSOURCE_0]
            edge_identity = Pipeline.Mv.Edge.RAW_CODEC_0

        elif ui_entry_point.ptype == PTYPE_SPEED:
            ui_videosource_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.VIDEOSOURCE_0]
            edge_identity = Pipeline.Speed.Edge.RAW_CODEC_0

        else:
            raise NotImplemented

        edge_gid = edge_identity.gid
        spool_dir = get_spool_dir(ui_pid)

        # create main.conf
        main_conf_jnode = {
            "FileApi": {
                "Store": join_path(spool_dir, 'rc%s-rc%s-store' % (edge_gid, edge_gid)),
                "FileCountLow": 18,
                "FileCountHigh": 20,
                "AddSceneReact": {
                    "Dir": join_path(spool_dir, 'mg%s-rc%s-scene' % (edge_gid, edge_gid))
                },
                "SceneEvent": [
                    join_path(spool_dir, 'rc%s-lr%s-scene' % (edge_gid, edge_gid)),
                    join_path(spool_dir, 'rc%s-tr%s-scene' % (edge_gid, edge_gid)),
                    join_path(spool_dir, 'mg%s-mcgst%s-scene' % (edge_gid, edge_gid))
                ]
            },
            "codec": {
                "Scale": {
                    "MaxWidth": 0,
                    "MaxHeight": 0,
                    "Interpolation": 1
                },
                "Encode": {
                    "UsedThreadMax": 1
                }
            }
        }

        # create setupstore.conf
        edge_id = helpers.get_edge_id(edge_group, edge_identity)

        if ui_videosource_jnode['ro']['raw-format'] == 'MONO8T':
            setupstore_conf_text = get_setupstore_conf_text(edge_id, False)
        else:
            # enable gc
            if ui_videosource_jnode['rw']['videostream']['control']['profile'] == 'auto-70-150':
                main_conf_jnode['FileApi']['SceneEvent'].append(
                    join_path(spool_dir, 'mg%s-gc%s-scene' % (edge_gid, edge_gid)))
            setupstore_conf_text = get_setupstore_conf_text(edge_id)

        return edge_id, main_conf_jnode, setupstore_conf_text

    def serialize(self):
        edge_id, main_conf_jnode, setupstore_conf_text = self._id_main_host_setupstore

        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_RAWCODEC, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        main_conf_abspath = join_path(edge_jnode_dir, 'main.conf')
        with open(main_conf_abspath, 'w') as json_file:
            ujson.dump(main_conf_jnode, json_file, indent=2)

        # create setupstore.conf
        setupstore_conf_filepath = join_path(edge_jnode_dir, 'setupstore.conf')
        with open(setupstore_conf_filepath, 'w') as text_file:
            text_file.write(setupstore_conf_text)


class RcDeleter:
    def __init__(self, edge_mc_dirpath):
        """
        :type edge_mc_dirpath: str
        """
        self._edge_mc_dirpath = edge_mc_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: RcDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Mv.Edge.RAW_CODEC_0)
            edge_type = Pipeline.Mv.Edge.RAW_CODEC_0.type
            deleter = RcDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            return None

        elif ui_entry_point.ptype == PTYPE_LPR_RTSP:
            return None

        else:
            raise NotImplemented

    def run(self):
        helpers.silent_rm_r(self._edge_mc_dirpath)
