#!/usr/bin/python

import gps
import time
import jsonschema
import ujson

from edge import eventlog, path_helpers as ph
from edge.admin import reply_helpers as rh, ui_types as ui
from edge.com import fs

import edge_creator as creator
from ui_pipeline_set import get_pipeline_type


def _geodata_rt_conf(new_jnode):
    """
    :type new_jnode: dict
    :rtype: dict
    """
    mode = new_jnode['mode']
    if mode == 'manual':
        jnode = {
            "source": ["static", new_jnode['latitude'], new_jnode['longitude']]
        }
    elif mode == 'auto':
        jnode = {
            "source": ["gpsd", "localhost", 2947]
        }
    else:
        jnode = dict()

    return jnode


def _check_changed(old_jnode, new_jnode):
    if old_jnode['mode'] != new_jnode['mode']:
        return True

    if old_jnode['mode'] == 'manual':
        if old_jnode['latitude'] != new_jnode['latitude']:
            return True

        if old_jnode['longitude'] != new_jnode['longitude']:
            return True
    else:
        pass

    return False


def _handle_get(ui_pid, entry_point):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    :rtype: dict
    """
    ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
    ui_jnode = fs.load_json2(ui_jnode_filepath)

    return rh.reply_done(ui_jnode['rw'])


def _handle_put(ui_pid, entry_point, new_ui_jnode):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    :type new_ui_jnode: dict

    :rtype: dict
    """

    def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
        """
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: bool
        """
        if not _check_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            return False

        ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
        ui_identity_to_jnode_map[entry_point]['rt'] = _geodata_rt_conf(new_ui_jnode)

        return True

    try:
        creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)

        return rh.reply_done(new_ui_jnode)

    except ui.LockError:
        return rh.reply_fail(rh.REPLY_CONFLICT_409)

    except Exception, x:
        eventlog.log_d("fail: %s" % x.message)
        return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)


def _get_gps_geodata(localhost, port):
    """
    :type localhost: str
    :type port: str
    :rtype: dict
    """
    session = gps.gps(localhost, port)
    session.stream(gps.WATCH_ENABLE | gps.WATCH_NEWSTYLE)

    result_jnode = None
    start_time = int(time.time())
    while True:
        try:
            report = session.next()
            if report['class'] == 'DEVICES' and len(report['devices']) == 0:
                return rh.reply_fail(rh.REPLY_GATEWAY_TIMEOUT)

            if report['class'] == 'TPV':
                result_jnode = {"latitude": report['lat'], "longitude": report['lon']}
                break
            end_time = int(time.time())
            diff_time = end_time - start_time
            if diff_time > 2:  # 2 seconds
                return rh.reply_fail(rh.REPLY_GATEWAY_TIMEOUT)

        except KeyError:
            pass
        except StopIteration:
            break

    if not result_jnode:
        return rh.reply_fail(rh.REPLY_NOT_FOUND_404)

    return rh.reply_done(result_jnode)


class UiGeoLocation:
    def __init__(self, handler):
        """
        :type handler: object -> dict
        """
        self._handler = handler

    @staticmethod
    def _validate_module_action(raw_action):
        """
        :type raw_action: str
        :rtype: (str | None, str | None)
        """
        return raw_action == 'define'

    @staticmethod
    def _deserialize_geolocation(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiGeoLocation
        """
        http_method = query_args.http_method
        url_path_info = query_args.path_info
        url_query_string = query_args.query_string
        body = ujson.loads(query_args.raw_body) if query_args.raw_body else None

        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiGeoLocation(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiGeoLocation(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        if http_method == 'GET':
            if url_path_info and not url_query_string:
                return UiGeoLocation(lambda: _handle_get(ui_pid, entry_point))

        elif http_method == 'PUT':
            schema_abs_path = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            request_schema_jnode = fs.load_json2(schema_abs_path)
            if body:
                try:
                    jsonschema.validate(body, request_schema_jnode)
                except (ValueError, jsonschema.ValidationError):
                    return UiGeoLocation(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

                return UiGeoLocation(lambda: _handle_put(ui_pid, entry_point, body))

        else:
            return UiGeoLocation(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405))

        return UiGeoLocation(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

    @staticmethod
    def _deserialize_geolocation_action(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiGeoLocation
        """
        http_method = query_args.http_method
        url_path_info = query_args.path_info
        url_query_string = query_args.query_string

        try:
            action = query_args.get_path_info_item(ui.UI_PIPELINE_ACTION_INDEX)
        except IndexError:
            return UiGeoLocation(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        success = UiGeoLocation._validate_module_action(action)
        if not success:
            return UiGeoLocation(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        if http_method == 'GET':
            if url_path_info and action and not url_query_string:
                return UiGeoLocation(lambda: _get_gps_geodata("localhost", "2947"))

        else:
            return UiGeoLocation(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405))

        return UiGeoLocation(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiGeoLocation
        """
        #
        url_path_info = query_args.path_info

        if len(url_path_info) == 4:
            return UiGeoLocation._deserialize_geolocation(query_args)

        elif len(url_path_info) == 5:
            return UiGeoLocation._deserialize_geolocation_action(query_args)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler()
