import jsonschema
import ujson

from shapely.geometry import polygon

from edge import eventlog, path_helpers as ph, pipeline_helpers as oh
from edge.admin import reply_helpers as rh, ui_types as ui
from edge.com import fs

import edge_creator as creator
from ui_pipeline_set import get_pipeline_type


class UiLaneEnforcement:
    def __init__(self, handler):
        """
        :type handler: object -> dict
        """

        self._handler = handler

    @staticmethod
    def _handle_get(ui_pid, entry_point):
        """
        :type ui_pid: str
        :type entry_point: ph.UiIdentity
        :rtype: dict
        """
        vs_stat = oh.get_videsource_status(ui_pid, entry_point.id)
        if vs_stat is None:
            eventlog.log_e('fail: http503. invalid videostream <none>')
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503,
                error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        elif vs_stat != 0:
            eventlog.log_e('fail: http503. invalid videostream <{}>'.format(vs_stat))
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503,
                error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        ui_jnode_path = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
        ui_jnode = fs.load_json2(ui_jnode_path)
        ui_jnode_rw = ui_jnode.get('rw', dict())

        stream_wh = oh.get_videosource_image_wh(ui_pid=ui_pid, ui_id=entry_point.id)
        if stream_wh:
            ui_jnode_rw['stats'] = {
                'scene-wh': stream_wh
            }

        return rh.reply_done(ui_jnode_rw)

    @staticmethod
    def _handle_put(ui_pid, entry_point, new_ui_jnode):
        """
        :type ui_pid: str
        :type entry_point: ph.UiIdentity
        :type new_ui_jnode: dict

        :rtype: dict
        """

        vs_stat = oh.get_videsource_status(ui_pid, entry_point.id)
        if vs_stat is None:
            eventlog.log_d('fail: get videosource status')
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503, error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        elif vs_stat != 0:
            eventlog.log_d('fail: videosource status error')
            return rh.reply_fail2(
                rh.REPLY_SERVICE_UNAVAILABLE_503, error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
                error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            if ui_identity_to_jnode_map[entry_point]['rw'] == new_ui_jnode:
                return False

            ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
            return True

        try:
            creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)
            wh = oh.get_videosource_image_wh(ui_pid=ui_pid, ui_id=entry_point.id)
            if wh is not None:
                w, h = wh
                new_ui_jnode['stats'] = {
                    'scene-wh': [w, h]
                }

            return rh.reply_done(new_ui_jnode)

        except ui.LockError:
            return rh.reply_fail(rh.REPLY_CONFLICT_409)

    @staticmethod
    def _deserialize_get(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLaneEnforcement
        """
        #
        url_query_string = query_args.query_string
        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        #
        if url_query_string:
            return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        return UiLaneEnforcement(lambda: UiLaneEnforcement._handle_get(ui_pid, entry_point))

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLaneEnforcement
        """
        url_query_string = query_args.query_string
        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        #
        if url_query_string:
            return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load new-ui-jnode
        new_ui_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_d('fail: bad-request')
            return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load ui-schena-jnode
        try:
            ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = ujson.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_d("fail: load ui-schema-jnode (%s)" % x.message)
            return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500))

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_d("fail: validate new-ui-jnode (%s)" % x.message)
            return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        try:
            # validate self-intersection of lanes
            lanes_jnode = new_ui_jnode['args']['lanes']
            for lane_jnode in lanes_jnode:
                polymask = polygon.Polygon(lane_jnode["mask"])
                if not polymask.is_valid:
                    eventlog.log_d("fail: bad-request (invalid road-mask)")
                    return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))
        except(ValueError, IndexError):
            eventlog.log_d("fail: bad-request (road-mask contains less then 3 vertexes)")
            return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # all arguments are valid
        return UiLaneEnforcement(lambda: UiLaneEnforcement._handle_put(ui_pid, entry_point, new_ui_jnode))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLaneEnforcement
        """
        if query_args.http_method == 'GET':
            return UiLaneEnforcement._deserialize_get(query_args)

        elif query_args.http_method == 'PUT':
            return UiLaneEnforcement._deserialize_put(query_args)

        else:
            eventlog.log_d("fail: method-not-allowed")
            return UiLaneEnforcement(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405))

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler()
