import jsonschema
import ujson

from edge import eventlog, path_helpers as ph, pipeline_helpers as helpers
from edge.admin import reply_helpers as rh, ui_types as ui
from edge.types import FormatTemplate
from edge.com import fs

import edge_creator as creator
from ui_pipeline_set import (get_pipeline_type, add_detector_to_pipeline,
                             delete_detector_from_pipeline, rename_detector)


def _validate_format(jnode):
    """
    :type jnode: list
    :rtype: bool
    """
    key_to_true_map = {
        'speed': True,
        'datetime': True,
        'datetime-ms': True,
        'post-name': True,
        'geolocation': True,
        'plate': True,
        'lane': True,
        'vendor': True,
        'model': True,
        'serial': True
    }

    for row in jnode:
        for col in row['cols']:
            try:
                FormatTemplate(template=col['format']).substitute(key_to_true_map)
            except KeyError as x:
                eventlog.log_d('fail: invalid key (%s)' % x.message)
                return False

    return True


class UiParking:
    def __init__(self, handler):
        """
        :type handler: () -> dict
        """
        self._handler = handler

    @staticmethod
    def _handle_get(ui_pid, ui_jnode_path):
        """
        :type ui_pid: str
        :type ui_jnode_path: str
        :rtype: dict
        """
        ui_jnode = fs.load_json2(ui_jnode_path)
        ui_jnode_rw = ui_jnode.get('rw', dict())
        ui_jnode_rw['limits'] = helpers.get_detector_limits(ui_pid=ui_pid, videosource_ui_id='0')

        return rh.reply_done(ui_jnode_rw)

    @staticmethod
    def _handle_put(ui_pid, entry_point, new_ui_jnode):
        """
        :type ui_pid: str
        :type entry_point: ph.UiIdentity
        :type new_ui_jnode: dict
        :rtype: dict
        """
        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            if ui_identity_to_jnode_map[entry_point]['rw'] == new_ui_jnode:
                return False

            if new_ui_jnode['args']['name'] != ui_identity_to_jnode_map[entry_point]['rw']['args']['name']:
                rename_detector(ui_pid, entry_point.type, entry_point.id, new_ui_jnode['args']['name'])

            ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
            return True

        try:
            creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)
            new_ui_jnode['limits'] = helpers.get_detector_limits(ui_pid=ui_pid, videosource_ui_id='0')
            return rh.reply_done(new_ui_jnode)

        except ui.LockError:
            return rh.reply_fail(rh.REPLY_CONFLICT_409)

        except Exception as x:
            eventlog.log_d("fail: %s" % x.message)
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    @staticmethod
    def _handle_post(ui_pid, ui_type, name):
        """
        :type ui_pid: str
        :type ui_type: str
        :type name: str
        """
        # add new module pipeline.json file
        result = add_detector_to_pipeline(ui_pid, ui_type, name)
        if result:
            new_ui_id, new_module_list = result
            ui_ptype = get_pipeline_type(ui_pid)

            if ui_ptype is None:
                return UiParking(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

            entry_point = ph.UiIdentity(ui_ptype, ui_type, new_ui_id)

            # create ui jnode from default
            new_ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, new_ui_id)
            def_jnode_filepath = ph.get_ui_def_jnode_filepath(entry_point, None)
            fs.create_jnode_from_default(def_jnode_filepath, new_ui_jnode_filepath)

            # create schema symlink
            ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, entry_point.type, new_ui_id)
            fs.ln_schema(ui_schema_filepath, new_ui_jnode_filepath)

            # get image wh for pipeline/ui_pid/videosource/0
            image_wh = helpers.get_videosource_image_wh(ui_pid, '0')

            def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
                """
                :type ui_identity_to_jnode_map: dict of UiIdentity: dict
                :rtype: bool
                """
                ui_identity_to_jnode_map[entry_point]['rw']['args']['name'] = name

                if image_wh:
                    polygon_jnode = ui_identity_to_jnode_map[entry_point]['rw']['args']['scene']['area']
                    polygon_jnode['scene-wh'] = [image_wh[0], image_wh[1]]
                    polygon_jnode['mask'] = [
                        [0, 0],
                        [image_wh[0], 0],
                        [image_wh[0], image_wh[1]],
                        [0, image_wh[1]]
                    ]
                return True

            try:
                creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)
                return rh.reply_done(new_module_list)

            except ui.LockError:
                return rh.reply_fail(rh.REPLY_CONFLICT_409)

            except Exception as x:
                eventlog.log_d("fail: %s" % x.message)
                return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

        else:
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    @staticmethod
    def _handle_delete(ui_pid, entry_point):
        """
        :type ui_pid: str
        :type entry_point: ph.UiIdentity
        """

        # remove ui_jnode from disk
        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
        fs.silent_rm_r(ui_jnode_filepath)

        # remove ui_schema from disk
        ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, entry_point.type, entry_point.id)
        fs.silent_rm_r(ui_schema_filepath)

        # remove module from pipeline.json
        res = delete_detector_from_pipeline(ui_pid, entry_point.type, entry_point.id, '')

        try:
            creator.generate(ui_pid, entry_point, None)
            return rh.reply_done(res)

        except ui.LockError:
            return rh.reply_fail(rh.REPLY_CONFLICT_409)

        except Exception as x:
            eventlog.log_d("fail: %s" % x.message)
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    @staticmethod
    def _deserialize_post(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiParking
        """
        if query_args.query_string:
            eventlog.log_d('fail: bad-request')
            return UiParking(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
        except IndexError:
            return UiParking(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        args = ujson.loads(query_args.raw_body)
        ui_type = args.get('type')
        det_name = args.get('name')

        if ui_type != ph.UI_TYPE_PARKING or det_name is None:
            eventlog.log_d('fail: invalid params [%s]' % query_args.raw_body)
            return UiParking(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        return UiParking(lambda: UiParking._handle_post(ui_pid, ui_type, det_name))

    @staticmethod
    def _deserialize_delete(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiParking
        """
        if query_args.query_string or query_args.raw_body:
            eventlog.log_d('fail: bad-request')
            return UiParking(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiParking(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)

        if ui_ptype is None:
            return UiParking(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        return UiParking(lambda: UiParking._handle_delete(ui_pid, entry_point))

    @staticmethod
    def _deserialize_get(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiParking
        """
        if query_args.query_string or query_args.raw_body:
            eventlog.log_d('fail: bad-request')
            return UiParking(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiParking(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id)

        return UiParking(lambda: UiParking._handle_get(ui_pid, ui_jnode_filepath))

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiParking
        """
        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiParking(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)

        if ui_ptype is None:
            return UiParking(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        if query_args.query_string:
            eventlog.log_d('fail: bad-request')
            return UiParking(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load new-ui-jnode
        new_ui_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_d('fail: bad-request')
            return UiParking(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load ui-schena-jnode
        try:
            ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = ujson.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_d("fail: load ui-schema-jnode (%s)" % x.message)
            return UiParking(lambda: rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500))

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_d("fail: validate new-ui-jnode (%s)" % str(x.message))
            return UiParking(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # all arguments are valid
        return UiParking(lambda: UiParking._handle_put(ui_pid, entry_point, new_ui_jnode))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiParking
        """
        if query_args.http_method == 'GET':
            return UiParking._deserialize_get(query_args)

        elif query_args.http_method == 'PUT':
            return UiParking._deserialize_put(query_args)
        
        elif query_args.http_method == 'POST':
            return UiParking._deserialize_post(query_args)

        elif query_args.http_method == 'DELETE':
            return UiParking._deserialize_delete(query_args)

        else:
            eventlog.log_d("fail: method-not-allowed")
            return UiParking(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405))

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler()
