import jsonschema
import ujson

from edge import eventlog, path_helpers as ph, pipeline_helpers as oh
from edge.admin import reply_helpers as rh, ui_types as ui
from edge.types import FormatTemplate
from edge.com import fs

import edge_creator as creator
from ui_pipeline_set import get_pipeline_type, rename_detector


def _validate_format(jnode):
    """
    :type jnode: list
    :rtype: bool
    """
    key_to_true_map = {
        'speed': True,
        'datetime': True,
        'datetime-ms': True,
        'post-name': True,
        'geolocation': True,
        'plate': True,
        'lane': True,
        'dir': True,
        'video-speed': True,
        'radar-speed': True,
        'radar-speed2': True,
        'vendor': True,
        'model': True,
        'serial': True
    }

    for row in jnode:
        for col in row['cols']:
            try:
                FormatTemplate(template=col['format']).substitute(key_to_true_map)
            except KeyError as x:
                eventlog.log_d('fail: invalid key (%s)' % x.message)
                return False

    return True


class UiTransit:
    def __init__(self, handler):
        """
        :type handler: () -> dict
        """
        self._handler = handler

    @staticmethod
    def _handle_get(ui_pid, ui_jnode_file):
        """
        :type ui_pid: str
        :type ui_jnode_file: str
        :rtype: dict
        """
        ui_jnode = fs.load_json2(ui_jnode_file)
        ui_jnode_rw = ui_jnode.get('rw', dict())

        ui_jnode_rw['limits'] = oh.get_detector_limits(ui_pid=ui_pid, videosource_ui_id='0')
        return rh.reply_done(ui_jnode_rw)

    @staticmethod
    def _handle_put(ui_pid, entry_point, new_ui_jnode):
        """
        :type ui_pid: str
        :type entry_point: ph.UiIdentity
        :type new_ui_jnode: dict
        :rtype: dict
        """
        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            if ui_identity_to_jnode_map[entry_point]['rw'] == new_ui_jnode:
                return False

            if new_ui_jnode['args']['name'] != ui_identity_to_jnode_map[entry_point]['rw']['args']['name']:
                rename_detector(ui_pid, entry_point.type, entry_point.id, new_ui_jnode['args']['name'])

            ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
            return True

        try:
            creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)
            new_ui_jnode['limits'] = oh.get_detector_limits(ui_pid=ui_pid, videosource_ui_id='0')
            return rh.reply_done(new_ui_jnode)

        except ui.LockError:
            return rh.reply_fail(rh.REPLY_CONFLICT_409)

        except Exception as x:
            eventlog.log_d("fail: %s" % x.message)
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    @staticmethod
    def _deserialize_get(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiTransit
        """
        if query_args.query_string or query_args.raw_body:
            eventlog.log_d('fail: bad-request')
            return UiTransit(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiTransit(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id)

        return UiTransit(lambda: UiTransit._handle_get(ui_pid, ui_jnode_filepath))

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiTransit
        """
        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiTransit(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)

        if ui_ptype is None:
            return UiTransit(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        if query_args.query_string:
            eventlog.log_d('fail: bad-request')
            return UiTransit(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load new-ui-jnode
        new_ui_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_d('fail: bad-request')
            return UiTransit(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # load ui-schena-jnode
        try:
            ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = ujson.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_d("fail: load ui-schema-jnode (%s)" % x.message)
            return UiTransit(lambda: rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500))

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_d("fail: validate new-ui-jnode (%s)" % x.message)
            return UiTransit(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # validate "format" jnodes from json
        images_jnode = new_ui_jnode['args']['images']
        vehicle_body_footer_rows = images_jnode['vehicle']['footer']['rows']
        scene_alt_footer_rows = images_jnode['scene']['footer']['rows']

        if _validate_format(vehicle_body_footer_rows) is False or _validate_format(scene_alt_footer_rows) is False:
            return UiTransit(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        # all arguments are valid
        return UiTransit(lambda: UiTransit._handle_put(ui_pid, entry_point, new_ui_jnode))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiTransit
        """
        if query_args.http_method == 'GET':
            return UiTransit._deserialize_get(query_args)

        elif query_args.http_method == 'PUT':
            return UiTransit._deserialize_put(query_args)

        else:
            eventlog.log_d("fail: method-not-allowed")
            return UiTransit(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405))

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler()
