import jsonschema
import ujson

from edge import eventlog, path_helpers as ph, pipeline_helpers as oh
from edge.admin import reply_helpers as rh, ui_types as ui
from edge.com import fs


import edge_creator as creator
from ui_pipeline_set import get_pipeline_type
from copy import deepcopy


def _check_mv_changed(old_ui_jnode, new_ui_jnode):
    return old_ui_jnode['videostream'] != new_ui_jnode['videostream']


def _check_mv_resolution_changed(old_ui_jnode, new_ui_jnode):
    return old_ui_jnode['videostream']['resolution'] != new_ui_jnode['videostream']['resolution']


def _check_gc_mode_changed(old_ui_jnode, new_ui_jnode):
    old_gc_mode = old_ui_jnode['videostream']['control']['profile'] == 'auto-70-150'

    new_gc_mode = new_ui_jnode['videostream']['control']['profile'] == 'auto-70-150'

    return old_gc_mode != new_gc_mode


def _check_rtsp_link_changed(old_ui_jnode, new_ui_jnode):
    return old_ui_jnode['streaming']['url'] != new_ui_jnode['streaming']['url']


def _check_rtsp_changed(old_ui_jnode, new_ui_jnode):
    if old_ui_jnode['streaming'] != new_ui_jnode['streaming']:
        return True

    if old_ui_jnode['videostream']['hw-decode'] != new_ui_jnode['videostream']['hw-decode']:
        return True

    if old_ui_jnode['videostream']['codec'] != new_ui_jnode['videostream']['codec']:
        return True

    if old_ui_jnode['videostream']['color-mode'] != new_ui_jnode['videostream']['color-mode']:
        return True

    if old_ui_jnode['videostream']['SkipRate'] != new_ui_jnode['videostream']['SkipRate']:
        return True

    return False


def _handle_get(ui_pid, entry_point):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    """
    ui_jnode_path = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
    ui_jnode = fs.load_json2(ui_jnode_path)
    ui_jnode_rw = ui_jnode.get('rw', dict())

    if ui_jnode_rw['type'] == 'rtsp':
        image_wh = oh.get_videosource_image_wh(ui_pid, entry_point.id, ui_jnode)
        if image_wh:
            ui_jnode_rw['videostream']['resolution'] = {
                "w": image_wh[0],
                "h": image_wh[1]
            }

    return rh.reply_done(ui_jnode_rw)


def _handle_put(ui_pid, entry_point, new_ui_jnode):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    :type new_ui_jnode: dict

    :rtype: dict
    """
    if entry_point.ptype == ph.PTYPE_SPEED or entry_point.ptype == ph.PTYPE_LPR_MV:
        reply_jnode = _put_mv(ui_pid, entry_point, new_ui_jnode)

    elif entry_point.ptype == ph.PTYPE_LPR_RTSP:
        reply_jnode = _put_rtsp(ui_pid, entry_point, new_ui_jnode)

    else:
        eventlog.log_d("fail: unsupported pipeline type")
        reply_jnode = rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

    return reply_jnode


def _put_rtsp(ui_pid, entry_point, new_ui_jnode):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    :type new_ui_jnode: dict
    :rtype: dict
    """
    restart_blobstore = []

    def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
        """
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :return: dict
        """
        if not _check_rtsp_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            return False

        # if link changed:
        # restart blobstore with video archive to clear h264 buffer:
        if _check_rtsp_link_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            restart_blobstore.append(True)

        ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode

        return True

    try:
        creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)

        if restart_blobstore:
            edge_id = oh.get_edge_id(ui_pid, ph.EdgeIdentity(ph.EDGE_TYPE_BLOBSTORE, '0'))
            oh.sighup_edge_module(ph.EDGE_TYPE_BLOBSTORE, edge_id)

        return rh.reply_done(new_ui_jnode)

    except ui.LockError:
        return rh.reply_fail(rh.REPLY_CONFLICT_409)


def _put_mv(ui_pid, entry_point, new_ui_jnode):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    :type new_ui_jnode: dict
    :rtype: dict
    """
    restart_setupstore = []
    restart_blobstore = []
    if new_ui_jnode['videostream']['control']['profile'] in ['auto-127', 'auto-70-150']:
        min_gain = new_ui_jnode['videostream']['control']['gain-auto-min']
        max_gain = new_ui_jnode['videostream']['control']['gain-auto-max']

        min_exposure = new_ui_jnode['videostream']['control']['exposure-auto-min']
        max_exposure = new_ui_jnode['videostream']['control']['exposure-auto-max']

        # GrabControl works bad when min_value == max_value
        if min_gain >= max_gain or min_exposure >= max_exposure:
            eventlog.log_d('fail: min value bigger than max (gain or exposure)')
            return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

    def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
        """
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: bool
        """
        if not _check_mv_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            return False

        # if resolution changed:
        # restart blobstore with video archive to clear h264 buffer:
        if _check_mv_resolution_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            restart_blobstore.append(True)

        if _check_gc_mode_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            restart_setupstore.append(True)

        ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
        return True

    try:
        creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)

        if restart_setupstore:
            oh.sighup_setupstore()

        if restart_blobstore:
            edge_id = oh.get_edge_id(ui_pid, ph.EdgeIdentity(ph.EDGE_TYPE_BLOBSTORE, '0'))
            oh.sighup_edge_module(ph.EDGE_TYPE_BLOBSTORE, edge_id)

        return rh.reply_done(new_ui_jnode)

    except ui.LockError:
        return rh.reply_fail(rh.REPLY_CONFLICT_409)


class UiVideosource:
    def __init__(self, handler):
        """
        :type handler: object -> dict
        """
        self._handler = handler

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiVideosource
        """
        #
        http_method = query_args.http_method
        url_query_string = query_args.query_string
        body_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None

        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiVideosource(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiVideosource(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        #
        if url_query_string:
            eventlog.log_d('fail: wrong query-string')
            return UiVideosource(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        #
        if http_method == 'GET':
            if not body_jnode:
                return UiVideosource(lambda: _handle_get(ui_pid, entry_point))

        elif http_method == 'PUT':
            schema_abs_path = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            request_schema_jnode = fs.load_json2(schema_abs_path)
            if body_jnode:
                try:
                    jsonschema.validate(body_jnode, request_schema_jnode)
                    # ROI checks - (width_offset + ROI_width) < max_width
                    if entry_point.ptype == ph.PTYPE_SPEED or entry_point.ptype == ph.PTYPE_LPR_MV:
                        roi_check = deepcopy(body_jnode)
                        roi_check['videostream']['resolution']['w'] += roi_check['videostream']['resolution']['x']
                        roi_check['videostream']['resolution']['h'] += roi_check['videostream']['resolution']['y']
                        jsonschema.validate(roi_check, request_schema_jnode)

                    return UiVideosource(lambda: _handle_put(ui_pid, entry_point, body_jnode))

                except (ValueError, jsonschema.ValidationError) as x:
                    eventlog.log_d('fail: %s' % x.message)
                    return UiVideosource(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        else:
            eventlog.log_d('fail: method not allowed')
            return UiVideosource(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405))

        eventlog.log_d('fail: deserialize')
        return UiVideosource(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler()
