import os

from edge import eventlog
from edge.admin.reply_helpers import Reply, REPLY_NOT_FOUND_404, REPLY_SUCCESS_200
from edge.admin.cgi import (run_cgi, CgiModule, CgiPathInfo, CgiRequest,
                            CgiBadRequestError, CgiNotFoundError, CgiInternalServerError)


from _api import *


from edge import path_helpers as ph
from edge import pipeline_helpers as oh
from setupstore.actions.pipeline.edge_creator import generate, LockError


def _handle_regenerate():
    pipeline_jnode_set = oh.get_pipeline_jnode_set2()
    if pipeline_jnode_set is None:
        eventlog.log_d("fail: pipeline-jnode-set is absent")
        return

    for pipeline_jnode in pipeline_jnode_set:
        ui_pid = pipeline_jnode['id']
        ui_ptype = pipeline_jnode['type']

        eventlog.log_d('try: pipeline-%s' % ui_pid)
        # noinspection PyBroadException
        try:
            # create edge-jnode-set
            generate(ui_pid, ph.UiIdentity(ui_ptype, ph.UI_TYPE_SIGN, '0'), None)
            eventlog.log_d('done: pipeline-%s' % ui_pid)
            continue

        except NotImplementedError:
            eventlog.log_d('fail: pipeline-%s (camera not supported)' % ui_pid)

        except LockError:
            eventlog.log_d('fail: pipeline-%s (lock error)' % ui_pid)

        except:
            eventlog.log_d('fail: pipeline-%s (traceback)' % ui_pid)
            eventlog.log_x()

    eventlog.log_d('regeneration config done')


class _PathInfo(CgiPathInfo):
    def __init__(self, *args, **kwargs):
        CgiPathInfo.__init__(self, *args, **kwargs)

    @property
    def file_type(self):
        return self[0]

    @staticmethod
    def unpack(raw_path_info):
        """
        :type raw_path_info: str
        :rtype: _PathInfo
        """
        # eventlog.log_d('unpack %s' % type(raw_path_info))
        return _PathInfo(raw_path_info.split('/')[1:])


class _Module(CgiModule):
    """
    GET     /api/v2/admin/actions/sign
    PUT     /api/v2/admin/actions/sign
    PUT     /api/v2/admin/actions/sign/key
    PUT     /api/v2/admin/actions/sign/crt
    """
    def __init__(self, handler):
        """
        :type handler: () -> Reply
        """
        CgiModule.__init__(self, handler)

    @classmethod
    def unpack(cls, request):
        """
        :type request: CgiRequest
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        method_to_handlers_map = {
            'GET': [
                cls._get,
            ],
            'PUT': [
                cls._put,
                cls._put_key,
                cls._put_crt
            ],
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except CgiBadRequestError:
                raise

            except CgiNotFoundError:
                pass

        eventlog.log_i('fail: http404')
        raise CgiNotFoundError

    @staticmethod
    def _get(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 0:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """

            result = get()
            rc = REPLY_SUCCESS_200 if result else REPLY_NOT_FOUND_404
            return Reply(status=rc, body=result)

        return _Module(handler)

    @staticmethod
    def _put(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 0:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body is None:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            result = put(request.body_jnode)

            status = REPLY_NOT_FOUND_404
            if result:
                status = REPLY_SUCCESS_200

                # regenerate speedcontrol configs
                _handle_regenerate()

            return Reply(status=status, body=result)

        return _Module(handler)

    @staticmethod
    def _put_key(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 1:
                raise CgiNotFoundError

            if path_info.file_type != "key":
                raise CgiNotFoundError

        def check_args():
            """
            :rtype: str
            """
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            try:
                tmp_path = os.environ['BLOB_PATH']
            except KeyError:
                eventlog.log_e('fail: http500 (env.BLOB_PATH not found)')
                raise CgiInternalServerError

            return tmp_path

        check_path_info(request.path_info)
        key_filepath = check_args()

        def handler():
            """
            :rtype: Reply
            """
            reply = put_key(key_filepath)

            status = REPLY_NOT_FOUND_404
            if reply:
                status = REPLY_SUCCESS_200

                # restart speedcontrol
                oh.restart_edge_modules(ph.EDGE_TYPE_SPEEDCONTROL)

            return Reply(status=status, body=reply)

        return _Module(handler)

    @staticmethod
    def _put_crt(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 1:
                raise CgiNotFoundError

            if path_info.file_type != "crt":
                raise CgiNotFoundError

        def check_args():
            """
            :rtype: str
            """
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            try:
                tmp_path = os.environ['BLOB_PATH']
            except KeyError:
                eventlog.log_e('fail: http500 (env.BLOB_PATH not found)')
                raise CgiInternalServerError

            return tmp_path

        check_path_info(request.path_info)
        crt_filepath = check_args()

        def handler():
            """
            :rtype: Reply
            """
            reply = put_crt(crt_filepath)

            status = REPLY_NOT_FOUND_404
            if reply:
                status = REPLY_SUCCESS_200

                # restart speedcontrol
                oh.restart_edge_modules(ph.EDGE_TYPE_SPEEDCONTROL)

            return Reply(status=status, body=reply)

        return _Module(handler)


def _main():
    run_cgi(
        lambda arg: _PathInfo.unpack(arg),
        lambda arg: _Module.unpack(arg))


if __name__ == '__main__':
    eventlog.open_log('cgi-sign', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
