#!/usr/bin/python
import os
import ujson
import jsonschema

from edge import eventlog
from edge.admin import reply_helpers as reply
from edge.admin.cgi import (run_cgi2, CgiModule, CgiPathInfo, CgiLock,
                            CgiBackup, CgiInternalServerError,
                            CgiBadRequestError, CgiNotFoundError)

from edge.com.fs import ln_schema, create_jnode_from_default
from edge.com import sh


class _Assets:
    UI_SCHEMA = '/etc/edge/setupstore/schemas/zabbix.json'
    UI_DEF_JNODE = '/opt/edge/share/setupstore/ui/zabbix-jnode.json'
    UI_JNODE = '/etc/edge/setupstore/admin/actions/zabbix.json'

    WEB_CONF = '/etc/zabbix/zabbix_agentd.conf.d/web.conf'
    VARUSR_CONF = '/opt/varusr/local/etc/zabbix/varusr.conf'

    def __init__(self):
        raise NotImplementedError


class _Zabbix:
    def __init__(self):
        pass

    @staticmethod
    def get():
        """
        :rtype: dict

        :raise CgiInternalServerError:
        """
        def install_ui_jnode(def_jnode_path, jnode_path):
            """
            :type def_jnode_path: str
            :type jnode_path: str
            :raise CgiInternalServerError:
            """
            if os.path.exists(jnode_path):
                return

            # noinspection PyBroadException
            try:
                create_jnode_from_default(def_jnode_path, jnode_path)
            except Exception:
                eventlog.log_x()
                raise CgiInternalServerError

        def install_ui_schema(jnode_path, schema_path):
            """
            :type jnode_path: str
            :type schema_path: str
            """
            if os.path.islink(schema_path):
                if os.path.exists(os.readlink(schema_path)):
                    return

            # noinspection PyBroadException
            try:
                ln_schema(schema_path, jnode_path)
            except Exception:
                eventlog.log_x()
                raise CgiInternalServerError

        install_ui_jnode(_Assets.UI_DEF_JNODE, _Assets.UI_JNODE)
        install_ui_schema(_Assets.UI_JNODE, _Assets.UI_SCHEMA)

        # load ui-jnode
        try:
            with open(_Assets.UI_JNODE, 'r') as json_file:
                ui_jnode = ujson.load(json_file)
                return ui_jnode['rw']

        except (IOError, ValueError) as x:
            eventlog.log_e('fail: load {} (%s)'.format(_Assets.UI_JNODE, x.message))
            raise CgiInternalServerError

    @CgiLock('/var/run/edge/zabbix')
    @CgiBackup('/etc/edge/setupstore/backup/zabbix', {
        'zabbix.json': _Assets.UI_JNODE,
        'web.conf': _Assets.WEB_CONF,
    })
    def set(self, next_rw_jnode):
        """
        :type next_rw_jnode: dict

        :raise CgiInternalServerError:
        :raise CgiBadRequestError:
        """
        # load ui-schema
        try:
            with open(_Assets.UI_SCHEMA, 'r') as json_file:
                ui_schema_jnode = ujson.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_e("fail: load {} ({})".format(_Assets.UI_SCHEMA, x.message))
            raise CgiInternalServerError

        # validate ui-jnode against ui-schema
        try:
            jsonschema.validate(next_rw_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_e("fail: validate (%s)" % x.message)
            raise CgiBadRequestError

        # edit conf files
        with open(_Assets.UI_JNODE, 'r') as json_file:
            ui_jnode = ujson.load(json_file)

        if ui_jnode['rw'] == next_rw_jnode:
            return

        self._edit_web_conf(ui_jnode['rw'], next_rw_jnode)

        ui_jnode['rw'] = next_rw_jnode
        with open(_Assets.UI_JNODE, 'w') as json_file:
            ujson.dump(ui_jnode, json_file, indent=2)

        # restart services
        if next_rw_jnode['enabled']:
            sh.run_silent("update-rc.d zabbix-agent enable")
            sh.run_silent("service zabbix-agent restart")
        else:
            sh.run_silent("service zabbix-agent stop")
            sh.run_silent("update-rc.d zabbix-agent disable")

    @staticmethod
    def _edit_web_conf(last_rw_jnode, next_rw_jnode):
        def mk_header_lines():
            if os.path.exists(_Assets.WEB_CONF):
                with open(_Assets.WEB_CONF, 'r') as conf:
                    lines = conf.readlines()
            else:
                lines = []

            header_lines = []
            for line in lines:
                if line.startswith('#'):
                    header_lines.append(line)

            return header_lines

        conf_lines = mk_header_lines()

        enabled = next_rw_jnode['enabled']
        if enabled:
            conf_lines.append('\nListenPort={}\n'.format(next_rw_jnode['agent-port']))

            hostname = next_rw_jnode.get('agent-host')
            if hostname:
                conf_lines.append('\nHostname={}\n'.format(hostname))

            if not len(next_rw_jnode['servers']):
                next_rw_jnode['servers'] = last_rw_jnode['servers']

            conf_lines.append('Server=')
            servers = map(lambda server: server['host'], next_rw_jnode['servers'])
            conf_lines.append(', '.join(servers))

            conf_lines.append('\nServerActive=')
            servers = map(lambda server: ('%s:%d' % (server['host'], server['port'])), next_rw_jnode['servers'])
            conf_lines.append(', '.join(servers))

            conf_lines.append('\nInclude=/etc/edge/zabbix_agentd.d/\n')
            if os.path.exists(_Assets.VARUSR_CONF):
                conf_lines.append('\nInclude=%s\n' % _Assets.VARUSR_CONF)

        with open(_Assets.WEB_CONF, 'w') as web_conf:
            web_conf.write(''.join(conf_lines))


class _Module(CgiModule):
    """
    GET  /api/v2/admin/actions/zabbix  /
    PUT  /api/v2/admin/actions/zabbix  /
    """
    def __init__(self, handler):
        """
        :type handler: () -> Reply
        """
        CgiModule.__init__(self, handler)

    @classmethod
    def unpack(cls, request, zabbix):
        """
        :type request: CgiRequest
        :type zabbix: _Zabbix
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        method_to_handlers_map = {
            'GET': [
                cls._get
            ],
            'PUT': [
                cls._put,
            ]
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request, zabbix)

            except CgiBadRequestError:
                raise

            except CgiNotFoundError:
                pass

        eventlog.log_d('fail: http404')
        raise CgiNotFoundError

    @staticmethod
    def _get(request, zabbix):
        """
        :type request: CgiRequest
        :type zabbix: _Zabbix
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: CgiPathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 0:
                eventlog.log_i('fail: http404')
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            reply_jnode = zabbix.get()
            return reply.Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _put(request, zabbix):
        """
        :type request: CgiRequest
        :type zabbix: _Zabbix
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: CgiPathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 0:
                eventlog.log_i('fail: http404')
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if not request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            zabbix.set(request.body_jnode)
            reply_jnode = zabbix.get()
            return reply.Reply(body=reply_jnode)

        return _Module(handler)


def _main():
    zabbix = _Zabbix()

    run_cgi2(
        lambda arg: CgiPathInfo.unpack(arg),
        lambda arg: _Module.unpack(arg, zabbix))


if __name__ == '__main__':
    eventlog.open_log('cgi-zabbix', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
