import os
import socket

import ujson

from edge import eventlog
from edge.queue import JsonFileQueue
from edge.image.types import MemBlob
from edge.service.about import IAboutService
from edge.service.tz import ITzService
from edge.com.bs import BsError

from .stats import EventCounters
from .types import EventServiceArgs
from ..model.types import Event
from ..services.bs import IBsService
from ..services.systest import ISystestService


class IEventService:
    def __init__(self):
        pass

    def emit(self, event_args):
        """
        :type event_args: Event
        """
        raise NotImplementedError


def init_service(counters, args, tz_service, bs_service, systest_service,
                 about_service):
    """
    :type counters: EventCounters
    :type args: EventServiceArgs
    :type tz_service: ITzService
    :type bs_service: IBsService
    :type systest_service: ISystestService
    :type about_service: IAboutService
    :rtype: IEventService
    """
    json_q_list = map(lambda json_q_args: JsonFileQueue(json_q_args), args.json_queue_list)
    return _EventService(counters, tz_service, bs_service, systest_service,
                         about_service, json_q_list)


# noinspection PyUnusedLocal
def fini_service(service):
    """
    :type service: IEventService | _EventService
    """
    pass


def _make_fact(event_jnode, args_blob_ref, sign):
    def add_arch_file(name, altref=None):
        arch_file = {
                'ref': name,
                'name': name
            }
        if altref:
            arch_file['altref'] = altref

        arch_files.append(arch_file)

    def analyze_measurements():
        measurements = event_jnode['measurements']

        for measurement in measurements:
            waypoints = measurement['waypoints']
            best_images = waypoints['best'].get('images', {})
            plate_images = best_images.get('license-plates', [])

            # thumb.jpeg
            thumb_im_name = best_images['thumb']['name']
            add_arch_file(thumb_im_name, 'thumb.jpeg')

            # scene.jpeg
            scene_im_name = best_images['scene']['name']
            add_arch_file(scene_im_name, 'scene.jpeg')

            # vehicle.jpeg
            vehicle_im_name = best_images['vehicle']['name']
            add_arch_file(vehicle_im_name, 'vehicle.jpeg')

            # plate.jpeg
            map(lambda plate_im: add_arch_file(plate_im['name'], 'plate.jpeg'),
                plate_images)

    def analyze_evidence():
        e_jnode = event_jnode.get('evidence', {})
        em_jnode = e_jnode.get('measurements')
        if not em_jnode:
            return

        emw_jnodes = em_jnode.get('waypoints', {})
        for waypoint in emw_jnodes.itervalues():
            if type(waypoint) != list:
                waypoint = [waypoint]

            for w_jnode in waypoint:
                wi_jnodes = w_jnode.get('images', {})
                for wi_jnode in wi_jnodes.itervalues():
                    add_arch_file(wi_jnode['name'])

        emv_jnodes = em_jnode.get('videos', [])
        for emv_jnode in emv_jnodes:
            add_arch_file(emv_jnode['name'])

    arch_files = []
    add_arch_file(args_blob_ref, 'args.json')
    if sign:
        sign_name = os.path.splitext(args_blob_ref)[0] + '.sign'
        cert_name = os.path.splitext(args_blob_ref)[0] + '.crt'
        add_arch_file(sign_name)
        add_arch_file(cert_name)
    analyze_measurements()
    analyze_evidence()

    jnode = {
        'args': event_jnode,
        'dist': arch_files
    }

    return jnode


class _EventService(IEventService):
    def __init__(self, counters, tz_service, bs_service, systest_service,
                 about_service, json_q_list):
        """
        :type counters: EventCounters
        :type tz_service: ITzService
        :type bs_service: IBsService
        :type systest_service: ISystestService
        :type about_service: IAboutService
        :type json_q_list: list[JsonFileQueue]
        """
        IEventService.__init__(self)

        self._counters = counters

        self._tz_service = tz_service
        self._bs_service = bs_service
        self._systest_service = systest_service
        self._json_q_list = json_q_list
        self._devices_jnode = self._get_devices(about_service)

    def _get_devices(self, about_service):
        device = {'hostname': socket.gethostname()}
        hw = about_service.hw_info
        if any(hw):
            device['hw'] = hw
        return [device]

    def emit(self, event_args):
        """
        :type event_args: Event
        """
        # noinspection PyBroadException
        try:
            map(lambda blob: self._bs_service.save(event_args.sender, blob), event_args.get_blobs())

            event_jnode = event_args.serialize(self._tz_service)
            self._systest_service.serialize(event_jnode)
            event_jnode['origin']['devices'] = self._devices_jnode

            event_blob = MemBlob('args', event_args.ts,
                                 ujson.dumps(event_jnode),
                                 file_ext='json')
            sign_enabled = self._bs_service.sign(event_blob)
            self._bs_service.save(event_args.sender, event_blob)
            event_jnode = _make_fact(event_jnode, event_blob.name, sign_enabled)
            map(lambda json_q: json_q.add_jnode(event_jnode, event_blob.name), self._json_q_list)
            self._counters.inc_fact()
        except BsError:
            eventlog.log_d('fail: save to bs skip event')

        except Exception:
            eventlog.log_x()
