import ctypes

from glob import glob
from edge import eventlog
from edge.types import BufferSizeError
from .types import SignServiceArgs


class SignError(RuntimeError):
    pass


class _SignSessionError(SignError):
    pass


class ISignService:
    def __init__(self):
        pass

    @property
    def cert(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def hash_type(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    def sign_blob(self, blob_buffer, sign_buffer):
        """
        :type blob_buffer: str
        :type sign_buffer: str

        :rtype: int
        :raise SignError:
        """
        raise NotImplementedError

    def hash_blob(self, blob_buffer, hash_buffer):
        """
        :type blob_buffer: str
        :type hash_buffer: str

        :rtype: int
        :raise SignError:
        """
        raise NotImplementedError


class _StubService(ISignService):
    def __init__(self):
        ISignService.__init__(self)

    @property
    def cert(self):
        """
        :rtype: str
        """
        return ''

    @property
    def hash_type(self):
        """
        :rtype: str
        """
        return ''

    def sign_blob(self, blob_buffer, sign_buffer):
        return 0

    def hash_blob(self, blob_buffer, hash_buffer):
        return 0

    def destroy(self):
        pass


def init_service(args):
    """
    :type args: SignServiceArgs | None
    :rtype: ISignService
    """
    return _SignService(args) if args else _StubService()


def fini_service(service):
    """
    :type service: ISignService | _SignService
    """
    service.destroy()


class _ApiType:
    LIBSIGN = 1
    LIBSIGNIIT = 2

    def __init__(self):
        raise NotImplementedError


class _Api:
    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 3
    LOG_D = 4
    LOG_T = 5

    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2

    _lib = ctypes.CDLL(glob('/opt/edge/lib/libsign.so')[0])

    # int
    # EsOpen(const char *aCfgFile, uint8_t aLogLevel, void **aSession);
    open = _lib.EsOpen
    open.restype = ctypes.c_int
    open.argtypes = (ctypes.c_char_p, ctypes.c_uint8, ctypes.POINTER(ctypes.c_voidp))

    # int
    # EsSign2(void *aSession, const uint8_t *aBlobData, uint32_t aBlobSize,
    #     uint8_t *aSignData, uint32_t *aSignSize);
    sign2 = _lib.EsSign2
    sign2.restype = ctypes.c_int
    sign2.argtypes = (ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                      ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

    # int
    # EsHash2 (void *aSession, const uint8_t *aBlobData, uint32_t aBlobSize,
    #     uint8_t *aHashData, uint32_t *aHashSize);
    hash2 = _lib.EsHash2
    hash2.restype = ctypes.c_int
    hash2.argtypes = (ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                      ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

    # void
    # EsClose (void *aSession);
    close = _lib.EsClose
    close.argtype = ctypes.c_voidp

    hash_type = 'md5'

    def __init__(self):
        pass

    @classmethod
    def log_level_name_to_code(cls, name):
        name_to_code = {
            'none': cls.LOG_N,
            'error': cls.LOG_E,
            'warn': cls.LOG_W,
            'info': cls.LOG_I,
            'debug': cls.LOG_D,
            'trace': cls.LOG_T
        }
        return name_to_code.get(name, cls.LOG_I)


class _IitApi:
    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 0x33  # 0x13 stdout, 0x24 syslog
    LOG_D = 0x34  # 0x14 stdout, 0x23 syslog
    LOG_T = 5

    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2

    _lib = ctypes.CDLL(glob('/opt/edge/lib/libsigniit.so')[0])

    # int
    # SiitOpen (const char *aCfgFile, uint8_t aLogLevel, void **aSession);
    open = _lib.SiitOpen
    open.restype = ctypes.c_int
    open.argtypes = (ctypes.c_char_p, ctypes.c_uint8, ctypes.POINTER(ctypes.c_voidp))

    # int
    # SiitSign2 (void *aSession, const uint8_t *aBlobData, uint32_t aBlobDataSize,
    #     uint8_t *anOutSignBuff, uint32_t *anOutSignBuffSize);
    sign2 = _lib.SiitSign2
    sign2.restype = ctypes.c_int
    sign2.argtypes = (ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                      ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

    # int
    # SiitHash2 (void *aSession, const uint8_t *aBlobData, uint32_t aBlobDataSize,
    #     uint8_t *anOutHashBuff, uint32_t *anOutHashBuffSize);
    hash2 = _lib.SiitHash2
    hash2.restype = ctypes.c_int
    hash2.argtypes = (ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                      ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

    # void
    # SiitClose (void *aSession);
    close = _lib.SiitClose
    close.argtype = ctypes.c_voidp

    hash_type = 'gost34311'

    def __init__(self):
        pass

    @classmethod
    def log_level_name_to_code(cls, name):
        name_to_code = {
            'none': cls.LOG_N,
            'error': cls.LOG_E,
            'warn': cls.LOG_W,
            'info': cls.LOG_I,
            'debug': cls.LOG_D,
            'trace': cls.LOG_T
        }
        return name_to_code.get(name, cls.LOG_I)


_API_NAME_TO_TYPE_MAP = {
    'openssl': _ApiType.LIBSIGN,
    'iit': _ApiType.LIBSIGNIIT
}

_API_TYPE_TO_API = {
    _ApiType.LIBSIGN: _Api,
    _ApiType.LIBSIGNIIT: _IitApi
}


def _check_session(fn):
    def proxy_fn(self, *args, **kwargs):
        """
        :type self: _Session
        :raise _SignSessionError:
        """
        if self.closed:
            eventlog.log_e('fail: invalid session')
            raise _SignSessionError

        return fn(self, *args, **kwargs)

    return proxy_fn


class _Session:
    def __init__(self, handle, api=None):
        """
        :type handle: ctypes.c_void_p | None
        """
        self._handle = handle
        self.closed = handle is None
        self.api = api

    @staticmethod
    def open(conf_filepath, log_level, api):
        """
        :type conf_filepath: str
        :type log_level: int
        :type api: _Api | _IitApi
        :rtype: _Session

        :raise VsSessionError:
        """
        handle = ctypes.c_void_p(0)
        if log_level is None:
            log_level = api.LOG_I
        rc = api.open(conf_filepath, log_level, handle)
        if rc != api.RC_DONE:
            eventlog.log_e('fail: Open <{}>'.format(rc))
            raise _SignSessionError

        return _Session(handle=handle, api=api)

    @_check_session
    def sign_buffer(self, blob_buffer, sign_buffer):
        """
        :type blob_buffer: str
        :param sign_buffer: ctypes.create_string_buffer
        :type sign_buffer: str

        :rtype: int
        :raise SignError:
        """
        blob_data = ctypes.cast(blob_buffer, ctypes.POINTER(ctypes.c_uint8))
        blob_size = ctypes.c_uint32(len(blob_buffer))

        sign_data = ctypes.cast(sign_buffer, ctypes.POINTER(ctypes.c_uint8))

        init_sign_size = len(sign_buffer)
        sign_size = ctypes.c_uint32(init_sign_size)
        p_sign_size = ctypes.pointer(sign_size)

        rc = self.api.sign2(self._handle, blob_data, blob_size, sign_data, p_sign_size)
        if rc != self.api.RC_DONE:
            eventlog.log_e('fail: Sign <{}>'.format(rc))
            raise SignError

        fini_sign_size = sign_size.value
        if fini_sign_size > init_sign_size:
            eventlog.log_e('fail: sign-buffer too small (min-size: {})'.format(fini_sign_size))
            raise BufferSizeError(fini_sign_size)

        return fini_sign_size

    @_check_session
    def hash_buffer(self, blob_buffer, hash_buffer):
        """
        :type blob_buffer: str
        :param hash_buffer: ctypes.create_string_buffer
        :type hash_buffer: str

        :rtype: int
        :raise SignError:
        """
        blob_data = ctypes.cast(blob_buffer, ctypes.POINTER(ctypes.c_uint8))
        blob_size = ctypes.c_uint32(len(blob_buffer))

        hash_data = ctypes.cast(hash_buffer, ctypes.POINTER(ctypes.c_uint8))

        init_hash_size = len(hash_buffer)
        hash_size = ctypes.c_uint32(init_hash_size)
        p_hash_size = ctypes.pointer(hash_size)

        rc = self.api.hash2(self._handle, blob_data, blob_size, hash_data, p_hash_size)
        if rc != self.api.RC_DONE:
            eventlog.log_e('fail: Hash (%s)' % rc)
            raise SignError

        fini_hash_size = hash_size.value
        if fini_hash_size > init_hash_size:
            eventlog.log_e('fail: hash-buffer too small (min-size: {})'.format(fini_hash_size))
            raise BufferSizeError(fini_hash_size)

        return fini_hash_size

    def close(self):
        self.api.close(self._handle)
        self.closed = True


def _reopen_session(fn):
    def proxy_fn(self, *args, **kwargs):
        """
        :type self: _SignService
        """
        try_count = 0
        while try_count <= 3:
            try_count += 1
            try:
                if self._session.closed:
                    self._session = _Session.open(*self._session_conf)

                return fn(self, *args, **kwargs)

            except _SignSessionError:
                self._session.close()

        raise SignError

    return proxy_fn


class _SignService(ISignService):
    def __init__(self, conf):
        """
        :type conf: SignServiceArgs
        """
        ISignService.__init__(self)
        assert conf.provider_name in _API_NAME_TO_TYPE_MAP.keys()

        provider_type = _API_NAME_TO_TYPE_MAP[conf.provider_name]
        provider = _API_TYPE_TO_API[provider_type]
        provider_log = provider.log_level_name_to_code(conf.log_level)

        self._hash_type = provider.hash_type
        self._session_conf = conf.provider_conf, provider_log, provider
        self._session = _Session(handle=None)
        self._cert = conf.cert

    @property
    def cert(self):
        """
        :rtype: str
        """
        return self._cert

    @property
    def hash_type(self):
        return self._hash_type

    def destroy(self):
        self._session.close()

    @_reopen_session
    def sign_blob(self, blob_buffer, sign_buffer):
        sign_size = self._session.sign_buffer(blob_buffer, sign_buffer)
        return sign_size

    @_reopen_session
    def hash_blob(self, blob_buffer, hash_buffer):
        hash_size = self._session.hash_buffer(blob_buffer, hash_buffer)
        return hash_size
